import { Configuration } from './utils';
import { EntityRepository, EntityValidator } from './entity';
import { UnitOfWork } from './unit-of-work';
import { FilterQuery, IDatabaseDriver } from './drivers/IDatabaseDriver';
import { EntityData, EntityName, IEntity, IEntityType, IPrimaryKey } from './decorators';
import { QueryBuilder } from './query';
import { Connection } from './connections/Connection';
import { QueryOrder } from './query';
export declare class EntityManager {
    readonly config: Configuration;
    private readonly driver;
    private readonly validator;
    private readonly repositoryMap;
    private readonly entityLoader;
    private readonly metadata;
    private readonly unitOfWork;
    private readonly entityFactory;
    constructor(config: Configuration, driver: IDatabaseDriver<Connection>);
    getDriver<D extends IDatabaseDriver<Connection> = IDatabaseDriver<Connection>>(): D;
    getConnection<C extends Connection = Connection>(): C;
    getRepository<T extends IEntityType<T>>(entityName: EntityName<T>): EntityRepository<T>;
    getValidator(): EntityValidator;
    createQueryBuilder(entityName: EntityName<IEntity>): QueryBuilder;
    find<T extends IEntityType<T>>(entityName: EntityName<T>, where?: FilterQuery<T>, options?: FindOptions): Promise<T[]>;
    find<T extends IEntityType<T>>(entityName: EntityName<T>, where?: FilterQuery<T>, populate?: string[], orderBy?: Record<string, QueryOrder>, limit?: number, offset?: number): Promise<T[]>;
    findOne<T extends IEntityType<T>>(entityName: EntityName<T>, where: FilterQuery<T> | IPrimaryKey, populate?: string[]): Promise<T | null>;
    beginTransaction(): Promise<void>;
    commit(): Promise<void>;
    rollback(): Promise<void>;
    transactional(cb: (em: EntityManager) => Promise<any>): Promise<any>;
    nativeInsert<T extends IEntityType<T>>(entityName: EntityName<T>, data: EntityData<T>): Promise<IPrimaryKey>;
    nativeUpdate<T extends IEntityType<T>>(entityName: EntityName<T>, where: FilterQuery<T>, data: EntityData<T>): Promise<number>;
    nativeDelete<T extends IEntityType<T>>(entityName: EntityName<T>, where: FilterQuery<T> | string | any): Promise<number>;
    /**
     * Shortcut to driver's aggregate method. Available in MongoDriver only.
     */
    aggregate(entityName: EntityName<IEntity>, pipeline: any[]): Promise<any[]>;
    merge<T extends IEntityType<T>>(entityName: EntityName<T>, data: EntityData<T>): T;
    /**
     * Creates new instance of given entity and populates it with given data
     */
    create<T extends IEntityType<T>>(entityName: EntityName<T>, data: EntityData<T>): T;
    /**
     * Gets a reference to the entity identified by the given type and identifier without actually loading it, if the entity is not yet loaded
     */
    getReference<T extends IEntityType<T>>(entityName: EntityName<T>, id: IPrimaryKey): T;
    count<T extends IEntityType<T>>(entityName: EntityName<T>, where: FilterQuery<T>): Promise<number>;
    persist(entity: IEntity | IEntity[], flush?: boolean): Promise<void>;
    persistAndFlush(entity: IEntity | IEntity[]): Promise<void>;
    persistLater(entity: IEntity | IEntity[]): void;
    remove<T extends IEntityType<T>>(entityName: EntityName<T>, where: T | any, flush?: boolean): Promise<number>;
    removeEntity(entity: IEntity, flush?: boolean): Promise<void>;
    removeAndFlush(entity: IEntity): Promise<void>;
    removeLater(entity: IEntity): void;
    /**
     * flush changes to database
     */
    flush(): Promise<void>;
    /**
     * clear identity map, detaching all entities
     */
    clear(): void;
    canPopulate(entityName: string | Function, property: string): boolean;
    fork(): EntityManager;
    getUnitOfWork(): UnitOfWork;
}
export interface FindOptions {
    populate?: string[];
    orderBy?: Record<string, QueryOrder>;
    limit?: number;
    offset?: number;
}
