"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("./utils");
const entity_1 = require("./entity");
const unit_of_work_1 = require("./unit-of-work");
const query_1 = require("./query");
const metadata_1 = require("./metadata");
class EntityManager {
    constructor(config, driver) {
        this.config = config;
        this.driver = driver;
        this.validator = new entity_1.EntityValidator(this.config.get('strict'));
        this.repositoryMap = {};
        this.entityLoader = new entity_1.EntityLoader(this);
        this.metadata = metadata_1.MetadataStorage.getMetadata();
        this.unitOfWork = new unit_of_work_1.UnitOfWork(this);
        this.entityFactory = new entity_1.EntityFactory(this.unitOfWork, this.driver, this.config);
    }
    getDriver() {
        return this.driver;
    }
    getConnection() {
        return this.driver.getConnection();
    }
    getRepository(entityName) {
        entityName = utils_1.Utils.className(entityName);
        if (!this.repositoryMap[entityName]) {
            const meta = this.metadata[entityName];
            const RepositoryClass = this.config.getRepositoryClass(meta.customRepository);
            this.repositoryMap[entityName] = new RepositoryClass(this, entityName);
        }
        return this.repositoryMap[entityName];
    }
    getValidator() {
        return this.validator;
    }
    createQueryBuilder(entityName) {
        entityName = utils_1.Utils.className(entityName);
        return new query_1.QueryBuilder(entityName, this.metadata, this.getConnection());
    }
    async find(entityName, where = {}, populate, orderBy, limit, offset) {
        entityName = utils_1.Utils.className(entityName);
        this.validator.validateParams(where);
        const options = utils_1.Utils.isObject(populate) ? populate : { populate, orderBy, limit, offset };
        const results = await this.driver.find(entityName, where, options.populate || [], options.orderBy || {}, options.limit, options.offset);
        if (results.length === 0) {
            return [];
        }
        const ret = [];
        for (const data of results) {
            const entity = this.merge(entityName, data);
            ret.push(entity);
        }
        await this.entityLoader.populate(entityName, ret, options.populate || []);
        return ret;
    }
    async findOne(entityName, where, populate = []) {
        entityName = utils_1.Utils.className(entityName);
        if (!where || (typeof where === 'object' && Object.keys(where).length === 0)) {
            throw new Error(`You cannot call 'EntityManager.findOne()' with empty 'where' parameter`);
        }
        let entity = this.getUnitOfWork().tryGetById(entityName, where);
        if (entity && entity.isInitialized()) {
            await this.entityLoader.populate(entityName, [entity], populate);
            return entity;
        }
        this.validator.validateParams(where);
        const data = await this.driver.findOne(entityName, where, populate);
        if (!data) {
            return null;
        }
        entity = this.merge(entityName, data);
        await this.entityLoader.populate(entityName, [entity], populate);
        return entity;
    }
    async beginTransaction() {
        await this.driver.beginTransaction();
    }
    async commit() {
        await this.driver.commit();
    }
    async rollback() {
        await this.driver.rollback();
    }
    async transactional(cb) {
        const em = this.fork();
        await em.getDriver().transactional(async () => {
            const ret = await cb(em);
            await em.flush();
            return ret;
        });
    }
    async nativeInsert(entityName, data) {
        entityName = utils_1.Utils.className(entityName);
        this.validator.validateParams(data, 'insert data');
        return this.driver.nativeInsert(entityName, data);
    }
    async nativeUpdate(entityName, where, data) {
        entityName = utils_1.Utils.className(entityName);
        this.validator.validateParams(data, 'update data');
        this.validator.validateParams(where, 'update condition');
        return this.driver.nativeUpdate(entityName, where, data);
    }
    async nativeDelete(entityName, where) {
        entityName = utils_1.Utils.className(entityName);
        this.validator.validateParams(where, 'delete condition');
        return this.driver.nativeDelete(entityName, where);
    }
    /**
     * Shortcut to driver's aggregate method. Available in MongoDriver only.
     */
    async aggregate(entityName, pipeline) {
        entityName = utils_1.Utils.className(entityName);
        return this.driver.aggregate(entityName, pipeline);
    }
    merge(entityName, data) {
        entityName = utils_1.Utils.className(entityName);
        if (!data || (!data.id && !data._id)) {
            throw new Error('You cannot merge entity without id!');
        }
        const entity = utils_1.Utils.isEntity(data) ? data : this.entityFactory.create(entityName, data, true);
        entity_1.EntityAssigner.assign(entity, data);
        this.getUnitOfWork().addToIdentityMap(entity);
        return entity;
    }
    /**
     * Creates new instance of given entity and populates it with given data
     */
    create(entityName, data) {
        entityName = utils_1.Utils.className(entityName);
        return this.entityFactory.create(entityName, data, false);
    }
    /**
     * Gets a reference to the entity identified by the given type and identifier without actually loading it, if the entity is not yet loaded
     */
    getReference(entityName, id) {
        entityName = utils_1.Utils.className(entityName);
        if (this.getUnitOfWork().getById(entityName, id)) {
            return this.getUnitOfWork().getById(entityName, id);
        }
        return this.entityFactory.createReference(entityName, id);
    }
    async count(entityName, where) {
        entityName = utils_1.Utils.className(entityName);
        this.validator.validateParams(where);
        return this.driver.count(entityName, where);
    }
    async persist(entity, flush = this.config.get('autoFlush')) {
        if (flush) {
            await this.persistAndFlush(entity);
        }
        else {
            this.persistLater(entity);
        }
    }
    async persistAndFlush(entity) {
        entity = Array.isArray(entity) ? entity : [entity];
        for (const ent of entity) {
            this.getUnitOfWork().persist(ent);
        }
        await this.flush();
    }
    persistLater(entity) {
        entity = Array.isArray(entity) ? entity : [entity];
        for (const ent of entity) {
            this.getUnitOfWork().persist(ent);
        }
    }
    async remove(entityName, where, flush = this.config.get('autoFlush')) {
        entityName = utils_1.Utils.className(entityName);
        if (utils_1.Utils.isEntity(where)) {
            await this.removeEntity(where, flush);
            return 1;
        }
        return this.nativeDelete(entityName, where);
    }
    async removeEntity(entity, flush = this.config.get('autoFlush')) {
        if (flush) {
            await this.removeAndFlush(entity);
        }
        else {
            this.removeLater(entity);
        }
    }
    async removeAndFlush(entity) {
        this.getUnitOfWork().remove(entity);
        await this.flush();
    }
    removeLater(entity) {
        this.getUnitOfWork().remove(entity);
    }
    /**
     * flush changes to database
     */
    async flush() {
        await this.getUnitOfWork().commit();
    }
    /**
     * clear identity map, detaching all entities
     */
    clear() {
        this.getUnitOfWork().clear();
    }
    canPopulate(entityName, property) {
        entityName = utils_1.Utils.className(entityName);
        const [p, ...parts] = property.split('.');
        const props = this.metadata[entityName].properties;
        const ret = p in props && props[p].reference !== entity_1.ReferenceType.SCALAR;
        if (!ret) {
            return false;
        }
        if (parts.length > 0) {
            return this.canPopulate(props[p].type, parts.join('.'));
        }
        return ret;
    }
    fork() {
        return new EntityManager(this.config, this.driver);
    }
    getUnitOfWork() {
        const em = utils_1.RequestContext.getEntityManager() || this;
        return em.unitOfWork;
    }
}
exports.EntityManager = EntityManager;
