"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const mongodb_1 = require("mongodb");
const Connection_1 = require("./Connection");
const utils_1 = require("../utils");
class MongoConnection extends Connection_1.Connection {
    async connect() {
        this.client = await mongodb_1.MongoClient.connect(this.config.getClientUrl(), { useNewUrlParser: true });
        this.db = this.client.db(this.config.get('dbName'));
    }
    async close(force) {
        return this.client.close(force);
    }
    async isConnected() {
        return this.client.isConnected();
    }
    getCollection(name) {
        return this.db.collection(name);
    }
    getDefaultClientUrl() {
        return 'mongodb://localhost:27017';
    }
    async execute(query) {
        throw new Error(`${this.constructor.name} does not support generic execute method`);
    }
    async find(collection, where, orderBy, limit, offset) {
        let query = `db.getCollection("${collection}").find`;
        where = this.convertObjectIds(where);
        const resultSet = this.getCollection(collection).find(where);
        const now = Date.now();
        query += `(${JSON.stringify(where)})`;
        if (orderBy && Object.keys(orderBy).length > 0) {
            query += `.sort(${JSON.stringify(orderBy)})`;
            resultSet.sort(orderBy);
        }
        if (limit !== undefined) {
            query += `.limit(${limit})`;
            resultSet.limit(limit);
        }
        if (offset !== undefined) {
            query += `.skip(${offset})`;
            resultSet.skip(offset);
        }
        const res = await resultSet.toArray();
        this.logQuery(`${query}.toArray(); [took ${Date.now() - now} ms]`);
        return res;
    }
    async insertOne(collection, data) {
        data = this.convertObjectIds(data);
        const now = Date.now();
        const res = await this.getCollection(collection).insertOne(data);
        const query = `db.getCollection("${collection}").insertOne(${JSON.stringify(data)});`;
        this.logQuery(`${query} [took ${Date.now() - now} ms]`);
        return res;
    }
    async updateMany(collection, where, data) {
        where = this.convertObjectIds(where);
        data = this.convertObjectIds(data);
        const now = Date.now();
        const res = await this.getCollection(collection).updateMany(where, { $set: data });
        const query = `db.getCollection("${collection}").updateMany(${JSON.stringify(where)}, { $set: ${JSON.stringify(data)} });`;
        this.logQuery(`${query} [took ${Date.now() - now} ms]`);
        return res;
    }
    async deleteMany(collection, where) {
        where = this.convertObjectIds(where);
        const res = await this.getCollection(collection).deleteMany(where);
        const query = `db.getCollection("${collection}").deleteMany()`;
        const now = Date.now();
        this.logQuery(`${query} [took ${Date.now() - now} ms]`);
        return res;
    }
    async aggregate(collection, pipeline) {
        const query = `db.getCollection("${collection}").aggregate(${JSON.stringify(pipeline)}).toArray();`;
        const now = Date.now();
        const res = this.getCollection(collection).aggregate(pipeline).toArray();
        this.logQuery(`${query} [took ${Date.now() - now} ms]`);
        return res;
    }
    async countDocuments(collection, where) {
        where = this.convertObjectIds(where);
        const query = `db.getCollection("${collection}").countDocuments(${JSON.stringify(where)})`;
        const now = Date.now();
        const res = await this.getCollection(collection).countDocuments(where);
        this.logQuery(`${query} [took ${Date.now() - now} ms]`);
        return res;
    }
    convertObjectIds(payload) {
        if (payload instanceof mongodb_1.ObjectID) {
            return payload;
        }
        if (utils_1.Utils.isString(payload) && payload.match(/^[0-9a-f]{24}$/i)) {
            return new mongodb_1.ObjectID(payload);
        }
        if (Array.isArray(payload)) {
            return payload.map((item) => this.convertObjectIds(item));
        }
        if (utils_1.Utils.isObject(payload)) {
            Object.keys(payload).forEach(k => {
                payload[k] = this.convertObjectIds(payload[k]);
            });
        }
        return payload;
    }
}
exports.MongoConnection = MongoConnection;
