"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const sqlite = require("sqlite");
const fs_1 = require("fs");
const Connection_1 = require("./Connection");
class SqliteConnection extends Connection_1.Connection {
    async connect() {
        this.connection = await sqlite.open(this.config.get('dbName'));
        await this.connection.exec('PRAGMA foreign_keys = ON');
    }
    async close(force) {
        await this.connection.close();
    }
    async isConnected() {
        return this.connection['driver']['open'];
    }
    async beginTransaction(savepoint) {
        await this.execute(savepoint ? `SAVEPOINT ${savepoint}` : 'BEGIN', [], 'run');
    }
    async commit(savepoint) {
        await this.execute(savepoint ? `RELEASE SAVEPOINT ${savepoint}` : 'COMMIT', [], 'run');
    }
    async rollback(savepoint) {
        await this.execute(savepoint ? `ROLLBACK TO SAVEPOINT ${savepoint}` : 'ROLLBACK', [], 'run');
    }
    getDefaultClientUrl() {
        return '';
    }
    async execute(query, params = [], method = 'all') {
        params = params.map(p => {
            if (p instanceof Date) {
                p = p.toISOString();
            }
            if (typeof p === 'boolean') {
                p = +p;
            }
            return p;
        });
        try {
            const now = Date.now();
            const statement = await this.connection.prepare(query);
            const res = await statement[method](...params);
            await statement.finalize();
            this.logQuery(query + ` [took ${Date.now() - now} ms]`);
            return this.transformResult(res, method);
        }
        catch (e) {
            e.message += `\n in query: ${query}`;
            if (params.length) {
                e.message += `\n with params: ${JSON.stringify(params)}`;
            }
            throw e;
        }
    }
    async loadFile(path) {
        const file = fs_1.readFileSync(path);
        await this.connection.exec(file.toString());
    }
    transformResult(res, method) {
        if (method === 'run') {
            return {
                affectedRows: res.changes,
                insertId: res.lastID,
            };
        }
        return res;
    }
}
exports.SqliteConnection = SqliteConnection;
