"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const DatabaseDriver_1 = require("./DatabaseDriver");
const entity_1 = require("../entity");
const query_1 = require("../query");
const utils_1 = require("../utils");
class AbstractSqlDriver extends DatabaseDriver_1.DatabaseDriver {
    async find(entityName, where, populate = [], orderBy = {}, limit, offset) {
        const qb = this.createQueryBuilder(entityName);
        qb.select('*').populate(populate).where(where).orderBy(orderBy);
        if (limit !== undefined) {
            qb.limit(limit, offset);
        }
        const res = await qb.execute('all');
        return res.map((r) => this.mapResult(r, this.metadata[entityName]));
    }
    async findOne(entityName, where, populate = []) {
        if (utils_1.Utils.isPrimaryKey(where)) {
            where = { id: where };
        }
        const qb = this.createQueryBuilder(entityName);
        const res = await qb.select('*').populate(populate).where(where).limit(1).execute('get');
        return this.mapResult(res, this.metadata[entityName]);
    }
    async count(entityName, where) {
        const qb = this.createQueryBuilder(entityName);
        const res = await qb.count('id', true).where(where).execute('get');
        return res.count;
    }
    async nativeInsert(entityName, data) {
        const collections = this.extractManyToMany(entityName, data);
        if (Object.keys(data).length === 0) {
            data.id = null;
        }
        const qb = this.createQueryBuilder(entityName);
        const res = await qb.insert(data).execute('run');
        await this.processManyToMany(entityName, res.insertId, collections);
        return res.insertId;
    }
    async nativeUpdate(entityName, where, data) {
        if (utils_1.Utils.isPrimaryKey(where)) {
            where = { id: where };
        }
        const collections = this.extractManyToMany(entityName, data);
        let res;
        if (Object.keys(data).length) {
            const qb = this.createQueryBuilder(entityName);
            res = await qb.update(data).where(where).execute('run');
        }
        await this.processManyToMany(entityName, utils_1.Utils.extractPK(data.id || where), collections);
        return res ? res.affectedRows : 0;
    }
    async nativeDelete(entityName, where) {
        if (utils_1.Utils.isPrimaryKey(where)) {
            where = { id: where };
        }
        const qb = this.createQueryBuilder(entityName);
        const res = await qb.delete(where).execute('run');
        return res.affectedRows;
    }
    createQueryBuilder(entityName) {
        return new query_1.QueryBuilder(entityName, this.metadata, this.connection);
    }
    extractManyToMany(entityName, data) {
        if (!this.metadata[entityName]) {
            return {};
        }
        const props = this.metadata[entityName].properties;
        const ret = {};
        for (const k of Object.keys(data)) {
            const prop = props[k];
            if (prop && prop.reference === entity_1.ReferenceType.MANY_TO_MANY) {
                ret[k] = data[k];
                delete data[k];
            }
        }
        return ret;
    }
    async processManyToMany(entityName, pk, collections) {
        const props = this.metadata[entityName].properties;
        const owners = Object.keys(collections).filter(k => props[k].owner);
        for (const k of owners) {
            const prop = props[k];
            const fk1 = prop.joinColumn;
            const fk2 = prop.inverseJoinColumn;
            const qb1 = this.createQueryBuilder(prop.pivotTable);
            await qb1.delete({ [fk1]: pk }).execute();
            for (const item of collections[k]) {
                const qb2 = this.createQueryBuilder(prop.pivotTable);
                await qb2.insert({ [fk1]: pk, [fk2]: item }).execute();
            }
        }
    }
}
exports.AbstractSqlDriver = AbstractSqlDriver;
