"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const metadata_1 = require("../metadata");
const utils_1 = require("../utils");
const query_1 = require("../query");
const naming_strategy_1 = require("../naming-strategy");
class DatabaseDriver {
    constructor(config) {
        this.config = config;
        this.metadata = metadata_1.MetadataStorage.getMetadata();
        this.logger = this.config.getLogger();
        this.transactionLevel = 0;
        this.transactionRolledBack = false;
    }
    async aggregate(entityName, pipeline) {
        throw new Error(`Aggregations are not supported by ${this.constructor.name} driver`);
    }
    async loadFromPivotTable(prop, owners) {
        if (!this.getConfig().usesPivotTable) {
            throw new Error(`${this.constructor.name} does not use pivot tables`);
        }
        const fk1 = prop.joinColumn;
        const fk2 = prop.inverseJoinColumn;
        const pivotTable = prop.owner ? prop.pivotTable : this.metadata[prop.type].properties[prop.mappedBy].pivotTable;
        const orderBy = { [`${pivotTable}.${this.metadata[pivotTable].primaryKey}`]: query_1.QueryOrder.ASC };
        const items = owners.length ? await this.find(prop.type, { [fk1]: { $in: owners } }, [pivotTable], orderBy) : [];
        const map = {};
        owners.forEach(owner => map['' + owner] = []);
        items.forEach((item) => {
            map['' + item[fk1]].push(item);
            delete item[fk1];
            delete item[fk2];
        });
        return map;
    }
    normalizePrimaryKey(data) {
        return data;
    }
    denormalizePrimaryKey(data) {
        return data;
    }
    mapResult(result, meta) {
        if (!result || !meta) {
            return result || null;
        }
        const ret = Object.assign({}, result);
        Object.values(meta.properties).forEach(prop => {
            if (prop.fieldName && prop.fieldName in ret) {
                utils_1.Utils.renameKey(ret, prop.fieldName, prop.name);
            }
        });
        return ret;
    }
    getConnection() {
        return this.connection;
    }
    async beginTransaction() {
        this.transactionLevel++;
        await this.runTransaction('beginTransaction');
    }
    async commit() {
        if (this.transactionRolledBack) {
            throw new Error('Transaction commit failed because the transaction has been marked for rollback only');
        }
        await this.runTransaction('commit');
        this.transactionLevel = Math.max(this.transactionLevel - 1, 0);
    }
    async rollback() {
        await this.runTransaction('rollback');
        if (this.transactionLevel === 1) {
            this.transactionRolledBack = false;
        }
        else if (!this.getConfig().supportsSavePoints) {
            this.transactionRolledBack = true;
        }
        this.transactionLevel = Math.max(this.transactionLevel - 1, 0);
    }
    async transactional(cb) {
        try {
            await this.beginTransaction();
            const ret = await cb();
            await this.commit();
            return ret;
        }
        catch (e) {
            await this.rollback();
            throw e;
        }
    }
    isInTransaction() {
        return this.transactionLevel > 0;
    }
    getConfig() {
        return {
            usesPivotTable: true,
            supportsTransactions: true,
            supportsSavePoints: false,
            namingStrategy: naming_strategy_1.UnderscoreNamingStrategy,
        };
    }
    async runTransaction(method) {
        if (this.transactionLevel === 1 || this.getConfig().supportsSavePoints) {
            const useSavepoint = this.transactionLevel !== 1 && this.getConfig().supportsSavePoints;
            await this.connection[method](useSavepoint ? this.getSavePointName() : undefined);
        }
    }
    getSavePointName() {
        return `${this.constructor.name}_${this.transactionLevel}`;
    }
}
exports.DatabaseDriver = DatabaseDriver;
