"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const mongodb_1 = require("mongodb");
const DatabaseDriver_1 = require("./DatabaseDriver");
const MongoConnection_1 = require("../connections/MongoConnection");
const utils_1 = require("../utils");
const naming_strategy_1 = require("../naming-strategy");
class MongoDriver extends DatabaseDriver_1.DatabaseDriver {
    constructor() {
        super(...arguments);
        this.connection = new MongoConnection_1.MongoConnection(this.config);
    }
    async find(entityName, where, populate, orderBy, limit, offset) {
        where = this.renameFields(entityName, where);
        const res = await this.connection.find(this.metadata[entityName].collection, where, orderBy, limit, offset);
        return res.map((r) => this.mapResult(r, this.metadata[entityName]));
    }
    async findOne(entityName, where, populate = []) {
        if (utils_1.Utils.isPrimaryKey(where)) {
            where = { _id: new mongodb_1.ObjectID(where) };
        }
        where = this.renameFields(entityName, where);
        const res = await this.connection.find(this.metadata[entityName].collection, where, {}, 1);
        return this.mapResult(res[0], this.metadata[entityName]);
    }
    async count(entityName, where) {
        where = this.renameFields(entityName, where);
        return this.connection.countDocuments(this.metadata[entityName].collection, where);
    }
    async nativeInsert(entityName, data) {
        data = this.renameFields(entityName, data);
        const res = await this.connection.insertOne(this.metadata[entityName].collection, data);
        return res.insertedId;
    }
    async nativeUpdate(entityName, where, data) {
        if (utils_1.Utils.isPrimaryKey(where)) {
            where = { _id: new mongodb_1.ObjectID(where) };
        }
        where = this.renameFields(entityName, where);
        data = this.renameFields(entityName, data);
        const res = await this.connection.updateMany(this.metadata[entityName].collection, where, data);
        return res.modifiedCount;
    }
    async nativeDelete(entityName, where) {
        if (utils_1.Utils.isPrimaryKey(where)) {
            where = { _id: new mongodb_1.ObjectID(where) };
        }
        where = this.renameFields(entityName, where);
        const res = await this.connection.deleteMany(this.metadata[entityName].collection, where);
        return res.deletedCount || 0;
    }
    async aggregate(entityName, pipeline) {
        return this.connection.aggregate(this.metadata[entityName].collection, pipeline);
    }
    normalizePrimaryKey(data) {
        if (data instanceof mongodb_1.ObjectID) {
            return data.toHexString();
        }
        return data;
    }
    denormalizePrimaryKey(data) {
        return new mongodb_1.ObjectID(data);
    }
    getConfig() {
        return {
            usesPivotTable: false,
            supportsTransactions: false,
            supportsSavePoints: false,
            namingStrategy: naming_strategy_1.MongoNamingStrategy,
        };
    }
    renameFields(entityName, data) {
        data = Object.assign({}, data); // copy first
        utils_1.Utils.renameKey(data, 'id', '_id');
        const meta = this.metadata[entityName];
        Object.keys(data).forEach(k => {
            if (meta && meta.properties[k]) {
                const prop = meta.properties[k];
                if (prop.fieldName) {
                    utils_1.Utils.renameKey(data, k, prop.fieldName);
                }
            }
        });
        return data;
    }
}
exports.MongoDriver = MongoDriver;
