"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ArrayCollection_1 = require("./ArrayCollection");
const enums_1 = require("./enums");
class Collection extends ArrayCollection_1.ArrayCollection {
    constructor(owner, items, initialized = true) {
        super(owner, items);
        this.initialized = false;
        this.dirty = false;
        this._populated = false;
        this.initialized = !!items || initialized;
    }
    getItems() {
        this.checkInitialized();
        return super.getItems();
    }
    add(...items) {
        this.modify('add', items);
    }
    set(items, initialize = false) {
        if (initialize) {
            this.initialized = true;
        }
        super.set(items);
        if (initialize) {
            this.dirty = false;
        }
    }
    remove(...items) {
        this.modify('remove', items);
    }
    contains(item) {
        this.checkInitialized();
        return super.contains(item);
    }
    count() {
        this.checkInitialized();
        return super.count();
    }
    isInitialized(fully = false) {
        if (fully) {
            return this.initialized && this.items.every(item => item.isInitialized());
        }
        return this.initialized;
    }
    shouldPopulate() {
        return this._populated;
    }
    populated(populated = true) {
        this._populated = populated;
    }
    isDirty() {
        return this.dirty;
    }
    setDirty(dirty = true) {
        this.dirty = dirty && this.property.owner; // set dirty flag only to owning side
    }
    async init(populate = []) {
        const em = this.owner.__em;
        if (!this.initialized && this.property.reference === enums_1.ReferenceType.MANY_TO_MANY && em.getDriver().getConfig().usesPivotTable) {
            const map = await em.getDriver().loadFromPivotTable(this.property, [this.owner.id]);
            this.set(map[this.owner.id].map(item => em.merge(this.property.type, item)), true);
            return this;
        }
        // do not make db call if we know we will get no results
        if (this.property.reference === enums_1.ReferenceType.MANY_TO_MANY && (this.property.owner || em.getDriver().getConfig().usesPivotTable) && this.items.length === 0) {
            this.initialized = true;
            this.dirty = false;
            this.populated();
            return this;
        }
        const cond = this.createCondition();
        const order = this.items.map(item => item.id);
        this.items.length = 0;
        const items = await em.find(this.property.type, cond, populate);
        this.reorderItems(items, order);
        this.items.push(...items);
        Object.assign(this, items);
        this.initialized = true;
        this.dirty = false;
        this.populated();
        return this;
    }
    createCondition() {
        const cond = {};
        if (this.property.reference === enums_1.ReferenceType.ONE_TO_MANY) {
            cond[this.property.fk] = this.owner.id;
        }
        else { // MANY_TO_MANY
            this.createManyToManyCondition(cond);
        }
        return cond;
    }
    createManyToManyCondition(cond) {
        if (this.property.owner || this.owner.__em.getDriver().getConfig().usesPivotTable) {
            cond.id = { $in: this.items.map(item => item.id) };
        }
        else {
            cond[this.property.mappedBy] = this.owner.id;
        }
    }
    modify(method, items) {
        this.checkInitialized();
        super[method](...items);
        this.setDirty();
    }
    checkInitialized() {
        if (!this.isInitialized()) {
            throw new Error(`Collection ${this.property.type}[] of entity ${this.owner.constructor.name}[${this.owner.id}] not initialized`);
        }
    }
    /**
     * re-orders items after searching with `$in` operator
     */
    reorderItems(items, order) {
        if (this.property.reference === enums_1.ReferenceType.MANY_TO_MANY && this.property.owner) {
            items.sort((a, b) => {
                return order.indexOf(a.id) - order.indexOf(b.id);
            });
        }
    }
}
exports.Collection = Collection;
