"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("../utils");
const metadata_1 = require("../metadata");
const enums_1 = require("./enums");
exports.SCALAR_TYPES = ['string', 'number', 'boolean', 'Date'];
class EntityFactory {
    constructor(unitOfWork, driver, config) {
        this.unitOfWork = unitOfWork;
        this.driver = driver;
        this.config = config;
        this.metadata = metadata_1.MetadataStorage.getMetadata();
        this.hydrator = this.config.getHydrator(this);
    }
    create(entityName, data, initialized = true) {
        entityName = utils_1.Utils.className(entityName);
        data = Object.assign({}, data);
        const meta = this.metadata[entityName];
        // normalize PK to `id: string`
        if (data.id || data._id) {
            data.id = this.driver.normalizePrimaryKey(data.id || data._id);
            delete data._id;
        }
        const entity = this.createEntity(data, meta);
        this.hydrator.hydrate(entity, meta, data);
        if (initialized) {
            delete entity.__initialized;
        }
        else {
            entity.__initialized = initialized;
        }
        return entity;
    }
    createEntity(data, meta) {
        const Entity = require(meta.path)[meta.name];
        if (!data.id) {
            const params = this.extractConstructorParams(meta, data);
            meta.constructorParams.forEach(prop => delete data[prop]);
            return new Entity(...params);
        }
        if (this.unitOfWork.getById(meta.name, data.id)) {
            return this.unitOfWork.getById(meta.name, data.id);
        }
        // creates new entity instance, with possibility to bypass constructor call when instancing already persisted entity
        const entity = Object.create(Entity.prototype);
        entity.id = data.id;
        this.unitOfWork.addToIdentityMap(entity);
        return entity;
    }
    createReference(entityName, id) {
        entityName = utils_1.Utils.className(entityName);
        if (this.unitOfWork.getById(entityName, id)) {
            return this.unitOfWork.getById(entityName, id);
        }
        return this.create(entityName, { id }, false);
    }
    /**
     * returns parameters for entity constructor, creating references from plain ids
     */
    extractConstructorParams(meta, data) {
        return meta.constructorParams.map(k => {
            if (meta.properties[k].reference === enums_1.ReferenceType.MANY_TO_ONE && data[k]) {
                const entity = this.unitOfWork.getById(meta.properties[k].type, data[k]);
                return entity || this.createReference(meta.properties[k].type, data[k]);
            }
            return data[k];
        });
    }
}
exports.EntityFactory = EntityFactory;
