"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const metadata_1 = require("../metadata");
const enums_1 = require("./enums");
const utils_1 = require("../utils");
const Collection_1 = require("./Collection");
class EntityLoader {
    constructor(em) {
        this.em = em;
        this.metadata = metadata_1.MetadataStorage.getMetadata();
        this.driver = this.em.getDriver();
    }
    async populate(entityName, entities, populate, validate = true) {
        if (entities.length === 0) {
            return;
        }
        for (const field of populate) {
            if (validate && !this.em.canPopulate(entityName, field)) {
                throw new Error(`Entity '${entityName}' does not have property '${field}'`);
            }
            await this.populateField(entityName, entities, field);
        }
    }
    /**
     * preload everything in one call (this will update already existing references in IM)
     */
    async populateMany(entityName, entities, field) {
        // set populate flag
        entities.forEach(entity => {
            if (utils_1.Utils.isEntity(entity[field]) || entity[field] instanceof Collection_1.Collection) {
                entity[field].populated();
            }
        });
        const prop = this.metadata[entityName].properties[field];
        const filtered = this.filterCollections(entities, field);
        if (prop.reference === enums_1.ReferenceType.MANY_TO_MANY && this.driver.getConfig().usesPivotTable) {
            return this.findChildrenFromPivotTable(filtered, prop, field);
        }
        const data = await this.findChildren(entities, prop);
        if (data.length === 0) {
            return [];
        }
        this.initializeCollections(filtered, prop, field, data);
        return data;
    }
    initializeCollections(filtered, prop, field, children) {
        if (prop.reference === enums_1.ReferenceType.ONE_TO_MANY) {
            this.initializeOneToMany(filtered, children, prop, field);
        }
        if (prop.reference === enums_1.ReferenceType.MANY_TO_MANY && !prop.owner && !this.driver.getConfig().usesPivotTable) {
            this.initializeManyToMany(filtered, children, prop, field);
        }
    }
    initializeOneToMany(filtered, children, prop, field) {
        for (const entity of filtered) {
            const items = children.filter(child => child[prop.fk] === entity);
            entity[field].set(items, true);
        }
    }
    initializeManyToMany(filtered, children, prop, field) {
        for (const entity of filtered) {
            const items = children.filter(child => child[prop.mappedBy].contains(entity));
            entity[field].set(items, true);
        }
    }
    async findChildren(entities, prop) {
        const children = this.getChildReferences(entities, prop);
        let fk = this.metadata[prop.type].primaryKey;
        if (prop.reference === enums_1.ReferenceType.ONE_TO_MANY) {
            fk = this.metadata[prop.type].properties[prop.fk].fieldName;
        }
        else if (prop.reference === enums_1.ReferenceType.MANY_TO_MANY && !prop.owner) {
            fk = this.metadata[prop.type].properties[prop.mappedBy].fieldName;
        }
        if (children.length === 0) {
            return [];
        }
        const ids = utils_1.Utils.unique(children.map(e => e.id));
        return this.em.find(prop.type, { [fk]: { $in: ids } });
    }
    async populateField(entityName, entities, field) {
        // nested populate
        if (field.includes('.')) {
            const [f, ...parts] = field.split('.');
            await this.populateMany(entityName, entities, f);
            const children = [];
            entities.forEach(entity => {
                if (utils_1.Utils.isEntity(entity[f])) {
                    children.push(entity[f]);
                }
                else if (entity[f] instanceof Collection_1.Collection) {
                    children.push(...entity[f].getItems());
                }
            });
            const filtered = utils_1.Utils.unique(children);
            const prop = this.metadata[entityName].properties[f];
            await this.populate(prop.type, filtered, [parts.join('.')], false);
        }
        else {
            await this.populateMany(entityName, entities, field);
        }
    }
    async findChildrenFromPivotTable(filtered, prop, field) {
        const map = await this.driver.loadFromPivotTable(prop, filtered.map(e => e.id));
        const children = [];
        for (const entity of filtered) {
            const items = map[entity.id].map(item => this.em.merge(prop.type, item));
            entity[field].set(items, true);
            children.push(...items);
        }
        return children;
    }
    getChildReferences(entities, prop) {
        const name = prop.name;
        const filtered = this.filterCollections(entities, name);
        const children = [];
        if (prop.reference === enums_1.ReferenceType.ONE_TO_MANY) {
            children.push(...filtered.map(e => e[name].owner));
        }
        else if (prop.reference === enums_1.ReferenceType.MANY_TO_MANY && prop.owner) {
            children.push(...filtered.reduce((a, b) => [...a, ...b[name].getItems()], []));
        }
        else if (prop.reference === enums_1.ReferenceType.MANY_TO_MANY) { // inversed side
            children.push(...filtered);
        }
        else { // MANY_TO_ONE
            children.push(...entities.filter(e => utils_1.Utils.isEntity(e[name]) && !e[name].isInitialized()).map(e => e[name]));
        }
        return children;
    }
    filterCollections(entities, field) {
        return entities.filter(e => e[field] instanceof Collection_1.Collection && !e[field].isInitialized(true));
    }
}
exports.EntityLoader = EntityLoader;
