"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("../utils");
const entity_1 = require("../entity");
class MetadataValidator {
    validateEntityDefinition(metadata, name) {
        const meta = metadata[name];
        // entities have PK
        if (!meta.primaryKey) {
            throw utils_1.ValidationError.fromMissingPrimaryKey(meta);
        }
        const references = Object.values(meta.properties).filter(prop => prop.reference !== entity_1.ReferenceType.SCALAR);
        for (const prop of references) {
            this.validateReference(meta, prop, metadata);
            if (prop.reference !== entity_1.ReferenceType.MANY_TO_ONE) {
                this.validateCollection(meta, prop, metadata);
            }
        }
    }
    validateReference(meta, prop, metadata) {
        // references do have types
        if (!prop.type) {
            throw utils_1.ValidationError.fromWrongTypeDefinition(meta, prop);
        }
        // references do have type of known entity
        if (!metadata[prop.type]) {
            throw utils_1.ValidationError.fromWrongTypeDefinition(meta, prop);
        }
    }
    validateCollection(meta, prop, metadata) {
        if (prop.reference === entity_1.ReferenceType.ONE_TO_MANY) {
            const owner = metadata[prop.type].properties[prop.fk];
            return this.validateOneToManyInverseSide(meta, prop, owner);
        }
        // m:n collection either is owner or has `mappedBy`
        if (!prop.owner && !prop.mappedBy && !prop.inversedBy) {
            throw utils_1.ValidationError.fromMissingOwnership(meta, prop);
        }
        if (prop.inversedBy) {
            const inverse = metadata[prop.type].properties[prop.inversedBy];
            this.validateManyToManyOwningSide(meta, prop, inverse);
        }
        else if (prop.mappedBy) {
            const inverse = metadata[prop.type].properties[prop.mappedBy];
            this.validateManyToManyInverseSide(meta, prop, inverse);
        }
    }
    validateOneToManyInverseSide(meta, prop, owner) {
        // 1:m collection has existing `fk` reference
        if (!owner) {
            throw utils_1.ValidationError.fromWrongReference(meta, prop, 'fk');
        }
        // 1:m collection has correct `fk` reference type
        if (owner.type !== meta.name) {
            throw utils_1.ValidationError.fromWrongReference(meta, prop, 'fk', owner);
        }
    }
    validateManyToManyOwningSide(meta, prop, inverse) {
        // m:n collection has correct `inversedBy` on owning side
        if (!inverse) {
            throw utils_1.ValidationError.fromWrongReference(meta, prop, 'inversedBy');
        }
        // m:n collection has correct `inversedBy` reference type
        if (inverse.type !== meta.name) {
            throw utils_1.ValidationError.fromWrongReference(meta, prop, 'inversedBy', inverse);
        }
        // m:n collection inversed side is not defined as owner
        if (inverse.inversedBy) {
            throw utils_1.ValidationError.fromWrongOwnership(meta, prop, 'inversedBy');
        }
    }
    validateManyToManyInverseSide(meta, prop, owner) {
        // m:n collection has correct `mappedBy` on inverse side
        if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY && prop.mappedBy && !owner) {
            throw utils_1.ValidationError.fromWrongReference(meta, prop, 'mappedBy');
        }
        // m:n collection has correct `mappedBy` reference type
        if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY && owner.type !== meta.name) {
            throw utils_1.ValidationError.fromWrongReference(meta, prop, 'mappedBy', owner);
        }
        // m:n collection owning side is not defined as inverse
        if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY && owner.mappedBy) {
            throw utils_1.ValidationError.fromWrongOwnership(meta, prop, 'mappedBy');
        }
    }
}
exports.MetadataValidator = MetadataValidator;
