"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("../utils");
const QueryBuilderHelper_1 = require("./QueryBuilderHelper");
const entity_1 = require("../entity");
const enums_1 = require("./enums");
/**
 * SQL query builder
 */
class QueryBuilder {
    constructor(entityName, metadata, connection) {
        this.entityName = entityName;
        this.metadata = metadata;
        this.connection = connection;
        this.aliasCounter = 1;
        this.flags = [];
        this._populate = {};
        this._leftJoins = {};
        this.alias = `e0`;
        this.helper = new QueryBuilderHelper_1.QueryBuilderHelper(this.entityName, this.alias, this.metadata);
    }
    select(fields) {
        this._fields = Array.isArray(fields) ? fields : [fields];
        return this.init(enums_1.QueryType.SELECT);
    }
    insert(data) {
        return this.init(enums_1.QueryType.INSERT, data);
    }
    update(data) {
        return this.init(enums_1.QueryType.UPDATE, data);
    }
    delete(cond = {}) {
        return this.init(enums_1.QueryType.DELETE, undefined, cond);
    }
    truncate() {
        return this.init(enums_1.QueryType.TRUNCATE);
    }
    count(field = 'id', distinct = false) {
        this.select(field);
        this.flags.push(enums_1.QueryFlag.COUNT);
        if (distinct) {
            this.flags.push(enums_1.QueryFlag.DISTINCT);
        }
        return this;
    }
    where(cond) {
        this._cond = this.processWhere(cond);
        return this;
    }
    orderBy(orderBy) {
        this._orderBy = orderBy;
        return this;
    }
    populate(populate) {
        populate.forEach(field => {
            if (this.metadata[field]) {
                const prop = this.metadata[field].properties[this.entityName];
                const alias = `e${this.aliasCounter++}`;
                this._leftJoins[field] = [this.metadata[field].collection, alias, prop.joinColumn, prop.inverseJoinColumn];
                this._populate[field] = alias;
            }
        });
        return this;
    }
    limit(limit, offset = 0) {
        this._limit = limit;
        if (offset) {
            this.offset(offset);
        }
        return this;
    }
    offset(offset) {
        this._offset = offset;
        return this;
    }
    getQuery() {
        let sql = this.getQueryBase();
        if (this._cond && Object.keys(this._cond).length > 0) {
            sql += ' WHERE ' + this.helper.getQueryCondition(this.type, this._cond).join(' AND ');
        }
        if (this._orderBy && Object.keys(this._orderBy).length > 0) {
            sql += ' ORDER BY ' + this.helper.getQueryOrder(this.type, this._orderBy, this._populate).join(', ');
        }
        sql += this.helper.getQueryPagination(this._limit, this._offset);
        return sql;
    }
    getParams() {
        let ret = [];
        if (this.type === enums_1.QueryType.INSERT && this._data) {
            ret = Object.values(this._data);
        }
        else if (this.type === enums_1.QueryType.UPDATE) {
            ret = Object.values(this._data);
        }
        if (this._cond) {
            ret = ret.concat(this.helper.getWhereParams(this._cond));
        }
        if (this._limit) {
            ret.push(this._limit);
        }
        if (this._offset) {
            ret.push(this._offset);
        }
        return ret;
    }
    async execute(method) {
        return this.connection.execute(this.getQuery(), this.getParams(), method);
    }
    prepareFields(fields, glue = ', ') {
        const ret = [];
        fields.forEach(f => {
            if (this._leftJoins[f]) {
                ret.push(...this.helper.mapJoinColumns(this.type, this._leftJoins[f]));
                return;
            }
            ret.push(this.helper.mapper(this.type, f));
        });
        Object.keys(this._populate).forEach(f => {
            ret.push(...this.helper.mapJoinColumns(this.type, this._leftJoins[f]));
            utils_1.Utils.renameKey(this._cond, this._leftJoins[f][3], `${this._leftJoins[f][1]}.${this._leftJoins[f][3]}`);
        });
        if (this.flags.includes(enums_1.QueryFlag.COUNT)) {
            if (this.flags.includes(enums_1.QueryFlag.DISTINCT)) {
                return `COUNT(DISTINCT ${ret[0]}) AS \`count\``;
            }
            return `COUNT(${ret[0]}) AS \`count\``;
        }
        return ret.join(glue);
    }
    processWhere(cond) {
        cond = Object.assign({}, cond); // copy first
        Object.keys(cond).forEach(field => {
            if (!this.metadata[this.entityName] || !this.metadata[this.entityName].properties[field]) {
                return;
            }
            const prop = this.metadata[this.entityName].properties[field];
            if (prop.reference === entity_1.ReferenceType.ONE_TO_MANY || prop.reference === entity_1.ReferenceType.MANY_TO_MANY) {
                this.processCollection(prop, cond, field);
            }
            else {
                utils_1.Utils.renameKey(cond, field, prop.fieldName);
            }
        });
        return cond;
    }
    processCollection(prop, cond, field) {
        if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY) {
            const alias1 = `e${this.aliasCounter++}`;
            if (prop.owner) {
                this._leftJoins[prop.name] = [prop.pivotTable, alias1, prop.joinColumn, prop.inverseJoinColumn];
            }
            else {
                const prop2 = this.metadata[prop.type].properties[prop.mappedBy];
                this._leftJoins[prop.name] = [prop2.pivotTable, alias1, prop.joinColumn, prop.inverseJoinColumn];
            }
            this._fields.push(prop.name);
            utils_1.Utils.renameKey(cond, field, `${alias1}.${prop.inverseJoinColumn}`);
        }
        else if (prop.reference === entity_1.ReferenceType.ONE_TO_MANY) {
            const alias2 = `e${this.aliasCounter++}`;
            const prop2 = this.metadata[prop.type].properties[prop.fk];
            this._leftJoins[prop.name] = [this.helper.getTableName(prop.type), alias2, prop2.fieldName, prop.referenceColumnName];
            utils_1.Utils.renameKey(cond, field, `${alias2}.${prop.referenceColumnName}`);
        }
    }
    init(type, data, cond) {
        this.type = type;
        if (data) {
            this._data = this.helper.processData(data);
        }
        if (cond) {
            this._cond = this.processWhere(cond);
        }
        return this;
    }
    getQueryBase() {
        let sql = this.type + ' ';
        switch (this.type) {
            case enums_1.QueryType.SELECT:
                sql += this.prepareFields(this._fields);
                sql += ` FROM \`${this.helper.getTableName(this.entityName)}\` AS \`${this.alias}\``;
                sql += this.helper.processJoins(this._leftJoins);
                break;
            case enums_1.QueryType.INSERT:
                sql += `INTO \`${this.helper.getTableName(this.entityName)}\``;
                sql += ' (' + Object.keys(this._data).map(k => this.helper.wrap(k)).join(', ') + ')';
                sql += ' VALUES (' + Object.keys(this._data).map(() => '?').join(', ') + ')';
                break;
            case enums_1.QueryType.UPDATE:
                sql += `\`${this.helper.getTableName(this.entityName)}\``;
                sql += ' SET ' + Object.keys(this._data).map(k => this.helper.wrap(k) + ' = ?').join(', ');
                break;
            case enums_1.QueryType.DELETE:
                sql += `FROM \`${this.helper.getTableName(this.entityName)}\``;
                break;
            case enums_1.QueryType.TRUNCATE:
                sql += `TABLE \`${this.helper.getTableName(this.entityName)}\``;
                break;
        }
        return sql;
    }
}
exports.QueryBuilder = QueryBuilder;
