"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("../utils");
const enums_1 = require("./enums");
class QueryBuilderHelper {
    constructor(entityName, alias, metadata) {
        this.entityName = entityName;
        this.alias = alias;
        this.metadata = metadata;
    }
    getWhereParams(conditions) {
        const ret = [];
        Object.values(conditions).forEach(cond => {
            if (utils_1.Utils.isObject(cond) && cond.$in) {
                return ret.push(...cond.$in);
            }
            else if (cond instanceof RegExp) {
                return ret.push(this.getRegExpParam(cond));
            }
            ret.push(cond);
        });
        return ret;
    }
    wrap(field) {
        if (field === '*') {
            return field;
        }
        return '`' + field + '`';
    }
    mapper(type, field, value = null) {
        let ret = this.wrap(field);
        if (field.match(/`?\w{2}`?\./)) {
            const [a, f] = field.split('.');
            ret = this.wrap(a) + '.' + this.wrap(f);
        }
        if (value) {
            ret += this.processValue(value);
        }
        if (type !== enums_1.QueryType.SELECT || ret.match(/`?\w{2}`?\./)) {
            return ret;
        }
        return this.wrap(this.alias) + '.' + ret;
    }
    processData(data) {
        data = Object.assign({}, data); // copy first
        Object.keys(data).forEach(k => {
            if (this.metadata[this.entityName] && this.metadata[this.entityName].properties[k]) {
                const prop = this.metadata[this.entityName].properties[k];
                if (prop.fieldName) {
                    utils_1.Utils.renameKey(data, k, prop.fieldName);
                }
            }
        });
        return data;
    }
    processJoins(leftJoins) {
        return Object.values(leftJoins).map(([table, alias, column]) => {
            return ` LEFT JOIN \`${table}\` AS \`${alias}\` ON \`${this.alias}\`.\`id\` = \`${alias}\`.\`${column}\``;
        }).join('');
    }
    mapJoinColumns(type, join) {
        return [
            this.mapper(type, `${join[1]}.${join[2]}`),
            this.mapper(type, `${join[1]}.${join[3]}`),
        ];
    }
    getTableName(entityName) {
        return this.metadata[entityName] ? this.metadata[entityName].collection : entityName;
    }
    getRegExpParam(re) {
        const value = re.source
            .replace(/\.\*/g, '%') // .* -> %
            .replace(/\./g, '_') // .  -> _
            .replace(/\\_/g, '.') // \. -> .
            .replace(/^\^/g, '') // remove ^ from start
            .replace(/\$$/g, ''); // remove $ from end
        if (re.source.startsWith('^') && re.source.endsWith('$')) {
            return value;
        }
        if (re.source.startsWith('^')) {
            return value + '%';
        }
        if (re.source.endsWith('$')) {
            return '%' + value;
        }
        return `%${value}%`;
    }
    getQueryCondition(type, cond) {
        return Object.keys(cond).map(k => this.mapper(type, k, cond[k]));
    }
    getQueryOrder(type, orderBy, populate) {
        return Object.keys(orderBy).map(k => {
            let alias = this.alias;
            let field = k;
            if (k.includes('.')) {
                [alias, field] = k.split('.');
            }
            alias = populate[alias] || alias;
            return this.mapper(type, `${alias}.${field}`) + ' ' + enums_1.QueryOrder[orderBy[k]];
        });
    }
    getQueryPagination(limit, offset) {
        let pagination = '';
        if (limit) {
            pagination += ' LIMIT ?';
        }
        if (offset) {
            pagination += ' OFFSET ?';
        }
        return pagination;
    }
    processValue(value) {
        if (value && utils_1.Utils.isObject(value) && value.$in) {
            return ` IN (${Object.keys(value.$in).map(() => '?').join(', ')})`;
        }
        if (value instanceof RegExp) {
            return ' LIKE ?';
        }
        if (value) {
            return ' = ?';
        }
    }
}
exports.QueryBuilderHelper = QueryBuilderHelper;
