"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const metadata_1 = require("../metadata");
const entity_1 = require("../entity");
const ChangeSetComputer_1 = require("./ChangeSetComputer");
const ChangeSetPersister_1 = require("./ChangeSetPersister");
const utils_1 = require("../utils");
class UnitOfWork {
    constructor(em) {
        this.em = em;
        /** map of references to managed entities */
        this.identityMap = {};
        /** holds copy of identity map so we can compute changes when persisting managed entities */
        this.originalEntityData = {};
        /** map of wrapped primary keys so we can compute change set without eager commit */
        this.identifierMap = {};
        this.persistStack = [];
        this.removeStack = [];
        this.changeSets = [];
        this.metadata = metadata_1.MetadataStorage.getMetadata();
        this.changeSetComputer = new ChangeSetComputer_1.ChangeSetComputer(this.em.getValidator(), this.originalEntityData, this.identifierMap);
        this.changeSetPersister = new ChangeSetPersister_1.ChangeSetPersister(this.em.getDriver(), this.identifierMap);
    }
    addToIdentityMap(entity) {
        this.identityMap[`${entity.constructor.name}-${entity.id}`] = entity;
        this.originalEntityData[entity.uuid] = utils_1.Utils.copy(entity);
    }
    getById(entityName, id) {
        const token = `${entityName}-${id}`;
        return this.identityMap[token];
    }
    tryGetById(entityName, where) {
        if (!utils_1.Utils.isPrimaryKey(where)) {
            return null;
        }
        where = this.em.getDriver().normalizePrimaryKey(where);
        return this.getById(entityName, where);
    }
    getIdentityMap() {
        return this.identityMap;
    }
    persist(entity, visited = []) {
        if (this.persistStack.includes(entity)) {
            return;
        }
        if (!entity.id) {
            this.identifierMap[entity.uuid] = new entity_1.EntityIdentifier();
        }
        this.persistStack.push(entity);
        this.cleanUpStack(this.removeStack, entity);
        this.cascade(entity, entity_1.Cascade.PERSIST, visited);
    }
    remove(entity, visited = []) {
        if (this.removeStack.includes(entity)) {
            return;
        }
        if (entity.id) {
            this.removeStack.push(entity);
        }
        this.cleanUpStack(this.persistStack, entity);
        this.unsetIdentity(entity);
        this.cascade(entity, entity_1.Cascade.REMOVE, visited);
    }
    async commit() {
        this.computeChangeSets();
        if (this.changeSets.length === 0) {
            return this.postCommitCleanup(); // nothing to do, do not start transaction
        }
        const driver = this.em.getDriver();
        const runInTransaction = !driver.isInTransaction() && driver.getConfig().supportsTransactions;
        const promise = utils_1.Utils.runSerial(this.changeSets, changeSet => this.commitChangeSet(changeSet));
        if (runInTransaction) {
            await driver.transactional(() => promise);
        }
        else {
            await promise;
        }
        this.postCommitCleanup();
    }
    clear() {
        Object.keys(this.identityMap).forEach(key => delete this.identityMap[key]);
        Object.keys(this.originalEntityData).forEach(key => delete this.originalEntityData[key]);
        this.postCommitCleanup();
    }
    unsetIdentity(entity) {
        delete this.identityMap[`${entity.constructor.name}-${entity.id}`];
        delete this.identifierMap[entity.uuid];
        delete this.originalEntityData[entity.uuid];
    }
    computeChangeSets() {
        this.changeSets.length = 0;
        while (this.persistStack.length) {
            this.findNewEntities(this.persistStack.shift());
        }
        for (const entity of Object.values(this.removeStack)) {
            const meta = this.metadata[entity.constructor.name];
            this.changeSets.push({ entity, delete: true, name: meta.name, collection: meta.collection, payload: {} });
        }
    }
    findNewEntities(entity) {
        const meta = this.metadata[entity.constructor.name];
        if (!entity.id && !this.identifierMap[entity.uuid]) {
            this.identifierMap[entity.uuid] = new entity_1.EntityIdentifier();
        }
        for (const prop of Object.values(meta.properties)) {
            const reference = entity[prop.name];
            this.processReference(prop, reference);
        }
        const changeSet = this.changeSetComputer.computeChangeSet(entity);
        if (changeSet) {
            this.changeSets.push(changeSet);
            this.cleanUpStack(this.persistStack, entity);
            this.originalEntityData[entity.uuid] = utils_1.Utils.copy(entity);
        }
    }
    processReference(prop, reference) {
        if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY && reference.isDirty()) {
            reference.getItems()
                .filter(item => !this.hasIdentifier(item))
                .forEach(item => this.findNewEntities(item));
        }
        else if (prop.reference === entity_1.ReferenceType.MANY_TO_ONE && reference && !this.hasIdentifier(reference)) {
            this.findNewEntities(reference);
        }
    }
    async commitChangeSet(changeSet) {
        const meta = this.metadata[changeSet.name];
        const pk = meta.primaryKey;
        const type = changeSet.entity[pk] ? (changeSet.delete ? 'Delete' : 'Update') : 'Create';
        await this.runHooks(`before${type}`, changeSet.entity, changeSet.payload);
        await this.changeSetPersister.persistToDatabase(changeSet);
        if (!changeSet.delete) {
            this.em.merge(changeSet.name, changeSet.entity);
        }
        await this.runHooks(`after${type}`, changeSet.entity);
    }
    async runHooks(type, entity, payload) {
        const hooks = this.metadata[entity.constructor.name].hooks;
        if (hooks && hooks[type] && hooks[type].length > 0) {
            const copy = utils_1.Utils.copy(entity);
            await utils_1.Utils.runSerial(hooks[type], hook => entity[hook]());
            if (payload) {
                Object.assign(payload, utils_1.Utils.diffEntities(copy, entity));
            }
        }
    }
    /**
     * clean up persist/remove stack from previous persist/remove calls for this entity done before flushing
     */
    cleanUpStack(stack, entity) {
        for (const index in stack) {
            if (stack[index] === entity) {
                stack.splice(+index, 1);
            }
        }
    }
    postCommitCleanup() {
        Object.keys(this.identifierMap).forEach(key => delete this.identifierMap[key]);
        this.persistStack.length = 0;
        this.removeStack.length = 0;
        this.changeSets.length = 0;
    }
    hasIdentifier(entity) {
        const pk = this.metadata[entity.constructor.name].primaryKey;
        if (entity[pk]) {
            return true;
        }
        return this.identifierMap[entity.uuid] && this.identifierMap[entity.uuid].getValue();
    }
    cascade(entity, type, visited) {
        if (visited.includes(entity)) {
            return;
        }
        visited.push(entity);
        switch (type) {
            case entity_1.Cascade.PERSIST:
                this.persist(entity, visited);
                break;
            case entity_1.Cascade.REMOVE:
                this.remove(entity, visited);
                break;
        }
        const meta = this.metadata[entity.constructor.name];
        for (const prop of Object.values(meta.properties)) {
            this.cascadeReference(entity, prop, type, visited);
        }
    }
    cascadeReference(entity, prop, type, visited) {
        if (!prop.cascade || !prop.cascade.includes(type)) {
            return;
        }
        if (prop.reference === entity_1.ReferenceType.MANY_TO_ONE && entity[prop.name]) {
            return this.cascade(entity[prop.name], type, visited);
        }
        const collection = entity[prop.name];
        if ([entity_1.ReferenceType.ONE_TO_MANY, entity_1.ReferenceType.MANY_TO_MANY].includes(prop.reference) && collection.isInitialized(true)) {
            collection.getItems().forEach(item => this.cascade(item, type, visited));
        }
    }
}
exports.UnitOfWork = UnitOfWork;
