"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const cache_1 = require("../cache");
const metadata_1 = require("../metadata");
const entity_1 = require("../entity");
const hydration_1 = require("../hydration");
const Logger_1 = require("./Logger");
const Utils_1 = require("./Utils");
class Configuration {
    constructor(options, validate = true) {
        this.cache = {};
        this.options = Utils_1.Utils.merge({}, Configuration.DEFAULTS, options);
        if (validate) {
            this.validateOptions();
        }
        this.logger = new Logger_1.Logger(this.options.logger, this.options.debug);
        this.driver = this.initDriver();
        this.init();
    }
    get(key, defaultValue) {
        return (this.options[key] || defaultValue);
    }
    getLogger() {
        return this.logger;
    }
    getClientUrl(hidePassword = false) {
        if (hidePassword) {
            return this.options.clientUrl.replace(/\/\/([^:]+):(\w+)@/, '//$1:*****@');
        }
        return this.options.clientUrl;
    }
    getDriver() {
        return this.driver;
    }
    getNamingStrategy() {
        return this.cached(this.options.namingStrategy || this.driver.getConfig().namingStrategy);
    }
    getHydrator(factory) {
        return this.cached(this.options.hydrator, factory, this.driver);
    }
    getMetadataProvider() {
        return this.cached(this.options.metadataProvider, this);
    }
    getCacheAdapter() {
        return this.cached(this.options.cache.adapter, this.options.cache.options);
    }
    getRepositoryClass(customRepository) {
        if (customRepository) {
            return customRepository();
        }
        else {
            return this.options.entityRepository;
        }
    }
    init() {
        if (!this.options.cache.enabled) {
            this.options.cache.adapter = cache_1.NullCacheAdapter;
        }
        if (!this.options.clientUrl) {
            this.options.clientUrl = this.driver.getConnection().getDefaultClientUrl();
        }
        if (this.options.entitiesDirsTs.length === 0) {
            this.options.entitiesDirsTs = this.options.entitiesDirs;
        }
    }
    validateOptions() {
        if (!this.options.dbName) {
            throw new Error('No database specified, please fill in `dbName` option');
        }
        if (this.options.entities.length === 0 && this.options.entitiesDirs.length === 0) {
            throw new Error('No entities found, please use `entities` or `entitiesDirs` option');
        }
    }
    initDriver() {
        if (!this.options.driver) {
            this.options.driver = require('../drivers/MongoDriver').MongoDriver;
        }
        return new this.options.driver(this);
    }
    cached(cls, ...args) {
        if (!this.cache[cls.name]) {
            const Class = cls;
            this.cache[cls.name] = new Class(...args);
        }
        return this.cache[cls.name];
    }
}
Configuration.DEFAULTS = {
    entities: [],
    entitiesDirs: [],
    entitiesDirsTs: [],
    tsConfigPath: process.cwd() + '/tsconfig.json',
    autoFlush: true,
    strict: false,
    logger: () => undefined,
    baseDir: process.cwd(),
    entityRepository: entity_1.EntityRepository,
    hydrator: hydration_1.ObjectHydrator,
    debug: false,
    cache: {
        enabled: true,
        adapter: cache_1.FileCacheAdapter,
        options: { cacheDir: process.cwd() + '/temp' },
    },
    metadataProvider: metadata_1.TypeScriptMetadataProvider,
};
exports.Configuration = Configuration;
