"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fastEqual = require("fast-deep-equal");
const clone = require("clone");
const metadata_1 = require("../metadata");
const entity_1 = require("../entity");
class Utils {
    static isObject(o) {
        return !!o && typeof o === 'object' && !Array.isArray(o);
    }
    static isString(s) {
        return typeof s === 'string';
    }
    static equals(a, b) {
        return fastEqual(a, b);
    }
    static unique(items) {
        return [...new Set(items)];
    }
    static merge(target, ...sources) {
        if (!sources.length) {
            return target;
        }
        const source = sources.shift();
        if (Utils.isObject(target) && Utils.isObject(source)) {
            Object.entries(source).forEach(([key, value]) => {
                if (Utils.isObject(value)) {
                    if (!target[key]) {
                        Object.assign(target, { [key]: {} });
                    }
                    Utils.merge(target[key], value);
                }
                else {
                    Object.assign(target, { [key]: value });
                }
            });
        }
        return Utils.merge(target, ...sources);
    }
    static diff(a, b) {
        const ret = {};
        Object.keys(b).forEach(k => {
            if (Utils.equals(a[k], b[k])) {
                return;
            }
            ret[k] = b[k];
        });
        return ret;
    }
    static diffEntities(a, b) {
        return Utils.diff(Utils.prepareEntity(a), Utils.prepareEntity(b));
    }
    static prepareEntity(e) {
        const metadata = metadata_1.MetadataStorage.getMetadata();
        const meta = metadata[e.constructor.name];
        const ret = Utils.copy(e);
        delete ret.__initialized;
        // remove collections and references
        Object.values(meta.properties).forEach(prop => {
            const pk = () => metadata[prop.type].primaryKey;
            const name = prop.name;
            if (e[name] instanceof entity_1.ArrayCollection || (Utils.isEntity(e[name]) && !e[name][pk()])) {
                return delete ret[name];
            }
            if (Utils.isEntity(e[name])) {
                return ret[prop.name] = ret[prop.name][pk()];
            }
        });
        // remove unknown properties
        Object.keys(e).forEach(prop => {
            if (!meta.properties[prop]) {
                delete ret[prop];
            }
        });
        return ret;
    }
    static copy(entity) {
        return clone(entity);
    }
    static renameKey(payload, from, to) {
        if (Utils.isObject(payload) && from in payload && !(to in payload)) {
            payload[to] = payload[from];
            delete payload[from];
        }
    }
    static getParamNames(func) {
        const STRIP_COMMENTS = /((\/\/.*$)|(\/\*[\s\S]*?\*\/))/mg;
        const ARGUMENT_NAMES = /([^\s,]+)/g;
        const fnStr = func.toString().replace(STRIP_COMMENTS, '');
        const result = fnStr.slice(fnStr.indexOf('(') + 1, fnStr.indexOf(')')).match(ARGUMENT_NAMES);
        if (result === null) {
            return [];
        }
        // handle class with no constructor
        if (result.length > 0 && result[0] === 'class') {
            return [];
        }
        // strip default values
        for (let i = 0; i < result.length; i++) {
            if (result[i].includes('=')) {
                result[i] = result[i].split('=')[0];
                result.splice(i + 1, 1);
            }
        }
        return result;
    }
    static isPrimaryKey(key) {
        return typeof key === 'string' || typeof key === 'number' || Utils.isObjectID(key);
    }
    static extractPK(data) {
        if (Utils.isPrimaryKey(data)) {
            return data;
        }
        if (Utils.isObject(data)) {
            return data.id || data._id || null;
        }
        return null;
    }
    static isEntity(data) {
        return Utils.isObject(data) && !!data.__entity;
    }
    static isObjectID(key) {
        return Utils.isObject(key) && key.constructor.name === 'ObjectID';
    }
    static className(classOrName) {
        if (typeof classOrName === 'string') {
            return classOrName;
        }
        return classOrName.name;
    }
    /**
     * uses some dark magic to get source path to caller where decorator is used
     */
    static lookupPathFromDecorator(meta) {
        if (meta.path) {
            return;
        }
        // use some dark magic to get source path to caller
        const stack = new Error().stack.split('\n');
        const line = stack.find(line => line.includes('__decorate'));
        if (line) {
            meta.path = line.match(/\((.*):\d+:\d+\)/)[1];
        }
    }
    static getObjectType(value) {
        const objectType = Object.prototype.toString.call(value);
        return objectType.match(/\[object (\w+)]/)[1].toLowerCase();
    }
    static async runSerial(items, cb) {
        const ret = [];
        for (const item of items) {
            ret.push(await cb(item));
        }
        return ret;
    }
}
exports.Utils = Utils;
